#!/usr/bin/env python3
# -*- coding: utf-8 -*-
'''
Created on Sun Apr 3 12:22:52 2016

You should execute this file to run and troubleshoot the homework
'''

import RLparent
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.animation as animation

from pylab import *
ion()

# %%


class CartPole(RLparent.RLparent):
    """
    Inverted pendulum / pole-cart / cart-pole reinforcement learning

         g=9.8      /
          |        / pole: Length = 1 m
          |       /
          V      /
                / θ (angle), theta_dot is angular velocity
         ______/_____
        |            | Cart: M = 1 kg
        |____________| ----> x_dot is velocity
          O        O
    L1--------x-------------------L2 x is poxition, with x limits of L1, L2)

    Actions: jerk left, jerk right (AKA bang-bang control)
    Goal: control x position of cart to keep pole close to upright,
    which is when θ = pi/2 (vertical); for visual see:
    https://upload.wikimedia.org/wikipedia/commons/9/9a/Degree-Radian_Conversion.svg
    """
    def BuildActionList(self):
        '''
        Actions of exerting horizontal force on cart
        '''
        return np.arange(-1.0, 1.1, 0.1)

    def BuildSateList(self):
        '''
        Initializes discrete state combinations
        '''
        s3div = (np.radians(15)-(np.radians(-15)))/8.0  # temp for x3

        s1 = [-1.0, 1.0]  # x position
        s2 = [-1.0, 0.0, 1.0]  # x_dot (velocity)
        s3 = np.arange(np.radians(-15), np.radians(15) + s3div, s3div)  # theta (angle)
        s4 = [-1.0, 1.0]  # theta dot (angular velocity)

        I, J, K, L = [np.size(_) for _ in [s1, s2, s3, s4]]

        states = []
        for i in range(I):
            for j in range(J):
                for k in range(K):
                    for l in range(L):
                        states.append([s1[i], s2[j], s3[k], s4[l]])
        return np.array(states)

    def GetReward(self, s):
        '''
        r: the returned reward
        f: true if goal/fail, otherwise f is false
        '''
        x, x_dot, theta, theta_dot = s

        r = 20 - 10*abs(10*theta)**2 - 5*abs(x) - 10*theta_dot
        f = False

        numDegrees = np.radians(45)  # can try various angles

        if x < -4.5 or x > 4.5 or theta < -numDegrees or theta > numDegrees:
            r = -100000 - 50*abs(x) - 100*abs(theta)
            f = True
        return r, f

    def DoAction(self, action, s):
        '''
        Takes an action (0 or 1) and
        the current values of the four state variables and
        updates values by estimating the state,
        Tau seconds later.
        The actual physics is here.
        '''
        x, x_dot, theta, theta_dot = s

        g = 9.8  # Gravity
        Mass_Cart = 1.0  # Mass of the cart is assumed to be 1Kg
        Mass_Pole = 0.1  # Mass of the pole is assumed to be 0.1Kg
        Total_Mass = Mass_Cart + Mass_Pole
        Length = 0.5  # Half of the length of the pole
        PoleMass_Length = Mass_Pole * Length
        Force_Mag = 10.0  # Force of push
        Tau = 0.02  # Time interval for updating the values
        Fourthirds = 4.0/3.0

        force = action * Force_Mag

        temp = (force + PoleMass_Length * theta_dot * theta_dot * np.sin(theta)) / Total_Mass

        thetaacc = (g * np.sin(theta) - np.cos(theta) * temp) / (Length * (Fourthirds - Mass_Pole * np.cos(theta) * np.cos(theta) / Total_Mass))

        xacc = temp - PoleMass_Length * thetaacc * np.cos(theta) / Total_Mass

        # Update the four state variables, using Euler's method:
        # https://en.wikipedia.org/wiki/Euler_method
        x = x + Tau * x_dot
        x_dot = x_dot + Tau * xacc
        theta = theta + Tau * theta_dot
        theta_dot = theta_dot + Tau * thetaacc
        xp = np.array([x, x_dot, theta, theta_dot])
        return xp

    def GetInitialState(self):
        '''
        Start state for every episode
        '''
        x = 0.0
        x_dot = -0.1
        theta = 0.0
        theta_dot = -0.1
        s = np.array([x, x_dot, theta, theta_dot])
        return s

# %%


def CartPoleDemo(maxepisodes, maxsteps=1000, Qmethod='SARSA',):
    '''
    Main function for running the example
    Qmethod: can be Q or SARSA
    '''
    CP = CartPole(alpha=0.3, gamma=.9, epsilon=0.1)
    runningStateHistory = []
    steps = 0
    maxSteps = 0
    for i in range(maxepisodes):

        # Both SARSA and Q are currently working:
        if Qmethod == 'Q':
            total_reward, steps, stateHistory = CP.QLearningEpisode(maxsteps)
        else:
            total_reward, steps, stateHistory = CP.SARSAEpisode(maxsteps)

        runningStateHistory.extend(stateHistory)

        print('Espisode: ', i,
              '|| Steps:', steps,
              '|| Reward:', str(total_reward),
              '|| epsilon: ', str(CP.epsilon))

        CP.epsilon = CP.epsilon * .99
        if steps > maxSteps:
            maxSteps = steps
        if steps == maxsteps:
            return runningStateHistory, stateHistory, maxSteps
    return runningStateHistory, stateHistory, maxSteps


def initGraph():
    '''
    Init for animated graph below
    '''
    line.set_data([], [])
    return line,


def animate(i):
    '''
    Each step/refresh of the animatd graph. This sort of gets "looped".
    '''
    thisx = [x1[i], x2[i]]
    thisy = [y1, y2[i]]
    line.set_data(thisx, thisy)
    return line,

# %%

if __name__ == '__main__':
    grandStateHistory, lastStateHistory, maxSteps = CartPoleDemo(maxepisodes=1000)

    # Just plotting from here down:
    toPlot = lastStateHistory  # or grandStateHistory for the whole training video

    # TOOD bonus: python updates broke these visualizations. Fix, and add a robot png image as the bottom dot, read guide with sublinks: https://stackoverflow.com/questions/22566284/matplotlib-how-to-plot-images-instead-of-points and plot horizontal line as "ground" below the robot image.

    xList = [state[0] for state in toPlot]
    thetaList = [state[2] for state in toPlot]
    x1 = xList
    y1 = 0
    x2 = 1*np.sin(thetaList) + x1
    y2 = 1*np.cos(thetaList) + y1

    fig = plt.figure()
    ax = plt.axes(xlim=(-4, 4), ylim=(-0.25, 1.25))
    ax.grid()
    line, = ax.plot([], [], 'o-', lw=2)
    plt.plot([-4, 4], [0, 0], lw=2)
    ani = animation.FuncAnimation(fig,
                                  animate,
                                  np.arange(1, len(xList)),
                                  interval=30,
                                  blit=False,
                                  init_func=initGraph,
                                  repeat=True)
