#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Apr 3 12:22:52 2016

ASSIGNMENT:
GOAL: Try to get the pole to stay up for at least 5,000 Steps (or more) in under 10,000 episodes, at least once or as many times as possible, when tested for 10 runs.

REQUIRED METHODS:
Add a rolling-calculated eligibility trace for each state as in TD(lambda) to update all the states weighted backwards with decaying importance, similar to Figure 11.14 in ArtInt (though using SARSA or Q-Learning). You should probably use a new self structure initialized in __init__.

OPTIONAL METHODS:
Optimize alpha, gamma, epsilon, reward function, or any other parts of the program (besides breaking it of course).

GRADING: You will be graded starting at 90 for completion, and then adding: (your max Step number in 50000 episodes divided by the max)*20. The format for auto-grading and import is bundled with this file (grading.py.) so you can test it out.

BONUS POINTS: You CAN go over 100% and this will help your grade with any or all of the following:

15 points - fix the 1 layer neural network (aka simple linear function) Q approximation to converge such that it reliably balances for > 500 or 1000 steps, or however long... If you can get it reliably working with a hidden layer nonlinear network, you get 30 points! Email the instructure if you complete this one.

10 points - Improve the graphics by plotting a moving robot png image just under the bottom dot, and a horiztonal line for the ground. If you do this, email the script and files to the instructor. I included a link showing how to do this for static plots, so it shouldn't be too hard.
"""

import numpy as np

# %%


class RLparent:
    '''
    Basic RL functions as a parent class to be inherited by each problem type
    INPUTS:
    1. alpha (learning rate)
    2. gamma (discount factor)
    3. epsilon (exploration probability)
    4. method = 'nn' for neural network, 'sa' for state-action
        (nn not currently converging)
    '''
    def __init__(self, alpha=0.3, gamma=.8, epsilon=0.01, method='sa'):
        self.method = method  # Q method
        self.alpha = alpha  # learning rate
        self.gamma = gamma  # discount factor
        self.epsilon = epsilon  # probability of a random action selection

        self.statelist = self.BuildSateList()  # the list of states
        self.actionlist = self.BuildActionList()  # the list of actions
        self.nactions = self.actionlist.shape[0]  # number of actions
        if method == 'sa':
            self.Q = self.BuildQTable()  # the Qtable
        else:
            self.Qnn = self.BuildQnnTable()  # the weight table for nn

    def __BuildActionList(self):
        """
        overwrite this function for your specific problem
        """
        # return a single array of real actions to be perfomed by the DoAction function
        # return actionlist

    def __BuildSateList(self):
        """
        overwrite this function for your specific problem
        """
        # return an array of vector states
        # this is typically a combinatorial discretization of the input space
        # return statelist

    def __GetReward(self, x):
        """
        overwrite this function for your specific problem
        """
        # r the current reward
        # f = True if the Goal is reached False otherwise
        # return r, f

    def __DoAction(self, force, x):
        """
        overwrite this function for your specific problem
        """
        # return xp, the new system state as an ND-vector
        # return xp

    def __GetInitialState(self):
        """
        overwrite this function for your specific problem
        """
        # return xp, the initial system state as an ND-vector
        # you could fix it or returing a random initial state
        # return x

    def BuildQTable(self):
        """
        Makes Q table: a fully-connected matrix of states by actions
        """
        nstates = self.statelist.shape[0]
        nactions = self.actionlist.shape[0]
        Q = [[0.0 for i in range(nactions)] for i in range(nstates)]
        return np.array(Q)

    def BuildQnnTable(self):
        """
        Makes weight vector table: a single-row vector of values for all features
        """
        nfeatures = self.statelist.shape[1] + 1  # 1 action
        Qnn = np.zeros((nfeatures, 1))
        return Qnn

    def DiscretizeState(self, x):
        """
        Checks which entry in the state list is more close to x,
        and returns the index of that entry.
        """
        edist = np.sqrt(np.sum((self.statelist-x)**2, 1))
        return np.argmin(edist)

    def e_greedy_selection(self, s):
        '''
        Selects an action using Epsilon-greedy strategy
        Q: the Qtable
        s: the current state
        '''
        if np.random.random() > self.epsilon:
            a = np.argmax(self.Q[s])
        else:
            # selects index of random action based on a uniform distribution
            a = np.random.randint(0, self.nactions)
        return a

    def e_greedy_selectionNN(self, xp):
        '''
        Selects an action using Epsilon-greedy strategy
        Q: the Qtable
        s: the current state
        '''
        if np.random.random() > self.epsilon:
            qMax = 0
            actionMax = 0
            for i in range(self.nactions):
                inputp = np.hstack((xp, self.actionlist[i]))
                qTemp = self.neuralNet(inputp, self.Qnn)
                if qTemp > qMax:
                    qMax = qTemp
                    actionMax = i
            a = actionMax
        else:
            # selects index of random action based on a uniform distribution
            a = np.random.randint(0, self.nactions)
        return a

    def UpdateSARSA(self, s, a, r, sp, ap):
        """
        Update Qtable using SARSA
        """
        self.Q[s][a] = self.Q[s][a] + self.alpha * (r + self.gamma * self.Q[sp][ap] - self.Q[s][a])

    def UpdateQ(self, s, a, r, sp):
        """
        Update Qvalues by QLearning
        """
        maxQ = max(self.Q[sp])
        self.Q[s][a] = self.Q[s][a] + self.alpha * (r + self.gamma * maxQ - self.Q[s][a])

    def UpdateQnn(self, x, a, r, xp, ap):
        '''
        Change feature values by: alpha * (reward + gamma * Q(xp,w) - Q(x,w)) * values
        NN could use fully continuous states (or not).
        Not currently working.
        '''
        # TODO bonus: Fix this (and/or associated functions) so it converges and learns well
        actionP = self.actionlist[ap]
        inputp = np.hstack((xp, actionP))
        Qp = self.neuralNet(inputp, self.Qnn)

        actionO = self.actionlist[a]
        inputo = np.hstack((x, actionO))
        Q = self.neuralNet(inputo, self.Qnn)

        # Transposing.T here is just to add each of the updated weights to the Q-feature table
        update = self.Qnn.T + self.alpha * (r + self.gamma * Qp - Q) * inputo
        self.Qnn = update.T

    def neuralNet(self, values, features):
        '''
        Single layer feed forward network with 1 output (Q).
        This is just a linear function approximation currently,
        but could be expanded.
        '''
        # return np.tanh(np.dot(values, features))
        return np.dot(values, features)

    def SARSAEpisode(self, maxsteps):
        '''
        Do one episode with SARSA learning
        maxstepts: the maximum number of steps per episode
        Q: the current QTable
        alpha: the current learning rate
        gamma: the current discount factor
        epsilon: probablity of a random action
        statelist: the list of states
        actionlist: the list of actions
        '''
        x = self.GetInitialState()
        steps = 0
        total_reward = 0
        stateHistory = []

        # convert the continous state variables to an index of the statelist
        s = self.DiscretizeState(x)

        # selects an action using the epsilon greedy selection strategy
        a = self.e_greedy_selection(s)

        for _ in range(1, maxsteps + 1):
            # convert the index of the action into an action value
            action = self.actionlist[a]

            # do the selected action and get the next car state
            xp = self.DoAction(action, x)

            # observe the reward at state xp and the final state flag
            r, isfinal = self.GetReward(xp)
            total_reward = total_reward + r

            # convert the continous state variables in [xp] to an index of the statelist
            sp = self.DiscretizeState(xp)

            if self.method == 'sa':
                # select action prime
                ap = self.e_greedy_selection(sp)

                # Update the Qtable, that is, learn from the experience
                self.UpdateSARSA(s, a, r, sp, ap)
            else:
                ap = self.e_greedy_selectionNN(xp)

                self.UpdateQnn(x, a, r, xp, ap)

            # update the current variables
            s = sp
            a = ap
            x = xp
            stateHistory.append(x)

            # increment the step counter.
            steps += 1

            # if it fails
            if isfinal is True:
                break
        return total_reward, steps, stateHistory

    def QLearningEpisode(self, maxsteps):
        """ does one episode of QLearning
        maxstepts: the maximum number of steps per episode
        alpha: the current learning rate
        gamma: the current discount factor
        epsilon: probablity of a random action
        actionlist: the list of actions
        """
        x = self.GetInitialState()
        steps = 0
        total_reward = 0
        stateHistory = []

        # convert the continous state variables to an index of the statelist
        s = self.DiscretizeState(x)

        for _ in range(1, maxsteps + 1):
            # selects an action using the epsilon greedy selection strategy
            a = self.e_greedy_selection(s)

            # convert the index of the action into an action value
            action = self.actionlist[a]

            # do the selected action and get the next car state
            xp = self.DoAction(action, x)
            sp = self.DiscretizeState(xp)

            # observe the reward at state xp and the final state flag
            r, isfinal = self.GetReward(xp)
            total_reward = total_reward + r

            # Update the Qtable, that is, learn from the experience
            self.UpdateQ(s, a, r, sp)

            # update the current variables
            s = sp
            x = xp
            stateHistory.append(x)

            # increment the step counter.
            steps += 1

            # if it fails
            if isfinal is True:
                break
        return total_reward, steps, stateHistory
